/********************************************************************
 *** pvswitch.c
 *** Copyright (c) 2004 by Markus Winand <mws@fatalmind.com>
 *** $Id: pvswitch.c,v 1.12 2004/06/17 21:27:48 mws Exp $
 *******************************************************************/
/*
    This file is part of pvswitch.

    pvswitch is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    pvswitch is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pvswitch; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <libgen.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "config.h"

#define PROGRAMNAME "pvswitch"
#define PROGRAMNAMEUP "PVSWITCH"
#define CONFIGFILE SYSCONFDIR "/" PROGRAMNAME ".conf"
#define TMP_BUFFER_SZ 1024


#define PART_ERROR      0

#define PART_PROGRAM    1
#define PART_APROGRAM   2
#define PART_KEY        3
#define PART_PATH       4
#define PART_COMMENT    5

#define PART_RECORDDONE 6

struct configrecord {
	char* part[PART_RECORDDONE];
	int  lineno; /* the line in the config file */
};



FILE* myfopen(const char *fn) {
	FILE* rv;
	rv = fopen(fn, "r");
	if (rv == NULL) {
		fprintf(stderr, PROGRAMNAME ": can't open %s: %s\n", fn, strerror(errno));
		exit(1);
	}
	return rv;
}

void strupper(char* str) {
	int i;
	for (i = 0; str[i] != 0; ++i) {
		str[i] = toupper(str[i]);	
	}
}


void skipTillEOL(FILE* fh) {
	int ch;
	while ((ch = fgetc(fh)) != EOF && (ch != '\n')) {
	}
	if (ch != EOF) {
		ungetc(ch, fh);
	}
}

void* mymalloc(size_t sz) {
	void* ptr = malloc(sz);
	if (ptr == NULL) {
		abort();
	}
	memset(ptr, 0, sz);
	return ptr;
}

void* myrealloc(void* buf, size_t old, size_t new) {
	void* ptr = realloc(buf, new);
	if (ptr == NULL) {
		abort();
	}
	if (new > old) {
		memset(ptr+old, 0, new-old);
	}
	return ptr;
}

void* myfree(void* ptr) {
	if (ptr != NULL) {
		free(ptr);
	}
	return NULL;
}

char* copyBufferSz(char* buffer, int sz) {
	char* ptr;

	buffer[sz++] = 0; /* terminate the string */
	ptr = mymalloc(sz + 1);
	strncpy(ptr, buffer, sz);

	return ptr;
}

char* copyBuffer(char* buffer) {
	char* ptr;
	ptr = mymalloc(strlen(buffer)+1);
	strcpy(ptr, buffer);
	return ptr;
}

char* mystat(char* fn) {
	struct stat s;
	if (stat(fn, &s) != 0) {
		char* buffer;

		buffer = mymalloc(strlen(fn) + TMP_BUFFER_SZ); /* should be enough */
		sprintf(buffer, "can't stat %s: %s\n", fn, strerror(errno));

		return buffer;
	} else {
		return NULL;
	}
}

void clearConfig(struct configrecord* cfg) {
	cfg->part[PART_PROGRAM]  = myfree(cfg->part[PART_PROGRAM] );
	cfg->part[PART_APROGRAM] = myfree(cfg->part[PART_APROGRAM]);
	cfg->part[PART_KEY]      = myfree(cfg->part[PART_KEY]     );
	cfg->part[PART_PATH]     = myfree(cfg->part[PART_PATH]    );
	cfg->part[PART_COMMENT]  = myfree(cfg->part[PART_COMMENT  ]);
}

void clearAllConfig(struct configrecord* cfg) {
	clearConfig(cfg);
	cfg->part[PART_ERROR]    = myfree(cfg->part[PART_ERROR    ]);
}

/* returns null on EOF */
struct configrecord* readConfigRecord(FILE* fh) {
	static int lineno = 1;		
	int ch;

	char buffer[TMP_BUFFER_SZ + 1];
	int bufidx = 0;
	int match  = 0;

	int part = PART_PROGRAM;
	int nxpart = part;
	struct configrecord* cfg = mymalloc(sizeof(struct configrecord));

	ch = fgetc(fh);
	while ((part != PART_RECORDDONE)) {
		int fieldterminator = 0;		
		if (((ch == '\n') && ((part != PART_PROGRAM) || (bufidx != 0))) || (ch == EOF)) {
			fieldterminator = 1;
			nxpart = PART_RECORDDONE;
		} else if ((ch == '\n')) {
			/* empty line */
			fieldterminator = 1;
			nxpart = PART_PROGRAM;
			++lineno;
		} else if ((part != PART_COMMENT) && (ch == ' ' || ch == '\t') && (bufidx != 0)) {
			fieldterminator = 1;
			if (part != PART_PROGRAM) {
				nxpart = part + 1;
			} else {
				nxpart = PART_KEY;
			}
		} else if ((part == PART_PROGRAM) && (ch == '#') && (bufidx == 0)) {
			/* complete line is comment */
			fieldterminator = 1;
			skipTillEOL(fh);	
			nxpart = PART_PROGRAM;
			++lineno;
		} else if ((part == PART_COMMENT) && (ch == '#')) {
			skipTillEOL(fh);
			fieldterminator = 1;
			nxpart = PART_RECORDDONE;
		} else if ((part == PART_PROGRAM) && (ch == '|')) {
			fieldterminator = 1;
			nxpart = PART_APROGRAM;
		}

		if (fieldterminator) {
			if (bufidx > 0) {
				cfg->part[part] = copyBufferSz(buffer, bufidx);
				if (part == PART_PROGRAM && nxpart == PART_KEY) {
					cfg->part[PART_APROGRAM] = copyBufferSz(buffer, bufidx);
				}
				bufidx = 0;
			}
		} else {
			if (!(bufidx == 0 && (ch == '\t' || ch == ' '))) { /* trim it */
				buffer[bufidx++] = ch;
				++match;
			} 
		}
		
		if (nxpart == PART_RECORDDONE) {
			cfg->lineno = lineno;
			++lineno;
			if (! (cfg->part[PART_PROGRAM ] != NULL 
				&& cfg->part[PART_APROGRAM] != NULL
				&& cfg->part[PART_KEY     ] != NULL
				&& cfg->part[PART_PATH    ] != NULL)
			) {
				clearConfig(cfg);
				cfg->part[PART_ERROR] = copyBuffer("mandatory field missing");
			}
		}
		part = nxpart;
		if (part != PART_RECORDDONE) {
			ch = fgetc(fh);
		}
	}
	/* if ((ch != EOF) || (cfg->part[PART_ERROR] == NULL)) { */
	if (ch == EOF && !match) {
		return myfree(cfg);
	} else {
		return cfg;
	}
}


struct configrecord** readConfig(char* program) {
	struct configrecord** cfg;
	struct configrecord* tmpcfg;
	int bufsz  = TMP_BUFFER_SZ;
	int bufidx = 0;
	FILE* fh;
	
	cfg = mymalloc(sizeof(struct configrecord) * bufsz);
	fh = myfopen(CONFIGFILE);

	while ((tmpcfg = readConfigRecord(fh)) != NULL) {
		if ((program == NULL) || (tmpcfg->part[PART_ERROR] != NULL) || (!strcmp(program, tmpcfg->part[PART_PROGRAM]))) {
			if ((bufidx+2) >= bufsz) { /* we alloc at least one more for the NULL */
				cfg = myrealloc(cfg, bufsz, (bufsz = bufsz + TMP_BUFFER_SZ));
			}
			cfg[bufidx++] = tmpcfg;
		}
	}
	cfg[bufidx] = NULL;
	if (bufidx == 0) {
		return myfree(cfg);
	} else {
		return cfg;
	}
}

void freeConfig(struct configrecord** cfg) {
	int i;
	for (i = 0; cfg[i] != NULL; ++i) {
		clearAllConfig(cfg[i]);
	}
	cfg = myfree(cfg);
}

int checkConfig(struct configrecord** cfg) {
	int i;
	int errors = 0;
	for (i = 0; cfg[i] != NULL; ++i) {
		if (cfg[i]->part[PART_ERROR] == NULL) {	
			cfg[i]->part[PART_ERROR] = mystat(cfg[i]->part[PART_PATH]);		
		}
		if (cfg[i]->part[PART_ERROR] != NULL) {	
			clearConfig(cfg[i]);
			++errors;
		}
	}
	return errors;
}

int printErrors(struct configrecord** cfg) {
	int errors = 0;
	int i;
	for (i = 0; cfg[i] != NULL; ++i) {
		if (cfg[i]->part[PART_ERROR] != NULL) {
			fprintf(stderr, CONFIGFILE ":%d  %s\n", cfg[i]->lineno, cfg[i]->part[PART_ERROR]);
			++errors;
		}
	}
	return errors;
}

struct configrecord* getConfig(struct configrecord** cfg, char* key) {
	int i;
	struct configrecord* rv = NULL;
	for (i = 0; (cfg[i] != NULL) && (rv == NULL); ++i) {
		if ((cfg[i]->part[PART_ERROR] == NULL) && ((key == NULL) || (!strcmp(cfg[i]->part[PART_KEY], key)))) {
			rv = cfg[i];
		}
	}
	return rv;	
}

void printConfig(struct configrecord** cfg) {
	int i;
	for (i = 0; (cfg[i] != NULL); ++i) {
		if (cfg[i]->part[PART_ERROR] == NULL) {
			if (cfg[i]->part[PART_COMMENT] != NULL) {
				fprintf(stderr, "%-15s  %s\n", cfg[i]->part[PART_KEY], cfg[i]->part[PART_COMMENT]);
			} else {
				fprintf(stderr, "%-15s  %s-%s\n", cfg[i]->part[PART_KEY], cfg[i]->part[PART_PROGRAM], cfg[i]->part[PART_KEY]);
			}
		}
	}
}



void printVersion(struct configrecord* cfg) {
	if (cfg != NULL) {
		printf("%s\n", cfg->part[PART_KEY]);
	} else {
		printf("\n");
	}
}

int strendcmp(char* end, char* ref) {
	int endlen;
	int reflen;

	endlen = strlen(end);
	reflen = strlen(ref);

	if (endlen > reflen) {
		return 0;
	}

	return !strcmp((ref + reflen - endlen), end);
}

void copynotice() {
        fprintf(stderr,
"\n"
"  %s V%s - program version switch\n"
"  Copyright (C) 2004 by Markus Winand <mws@fatalmind.com>\n"
"\n"
"  This program is free software; you can redistribute it and/or modify\n"
"  it under the terms of the GNU General Public License as published by\n"
"  the Free Software Foundation; either version 2 of the License, or\n"
"  (at your option) any later version.\n"
"\n"
"  This program is distributed in the hope that it will be useful,\n"
"  but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n"
"  GNU General Public License for more details.\n"
"\n"
"  You should have received a copy of the GNU General Public License\n"
"  along with this program; if not, write to the Free Software\n"
"  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n"
, PROGRAMNAME, VERSION);
}

char* getEnvVersion(char* bsnam) {
	char *ver;
	char *envname;
	char *bsnamup;
	bsnamup = copyBuffer(bsnam);

	strupper(bsnamup);
	envname = mymalloc(strlen(bsnam) + 25);
	sprintf(envname, PROGRAMNAMEUP "_%s_VERSION", bsnamup);
	free(bsnamup);
	ver = getenv(envname);
	free(envname);
	if (ver != NULL && ver[0] == 0) {
		ver = NULL;
	}
	return ver;
}

void configureVersion(struct configrecord** cfg, char* key) {
	struct configrecord* cfgrec;
	cfgrec = getConfig(cfg, key);
	if (cfgrec == NULL) {
		char* ver = getEnvVersion(cfg[0]->part[PART_APROGRAM]);
		fprintf(stderr, "%sver: no such %s version available\n", cfg[0]->part[PART_PROGRAM], cfg[0]->part[PART_PROGRAM]);
		if (ver != NULL) {
			cfgrec = getConfig(cfg, ver);
		}
	}
	printVersion(cfgrec);
}

char** myargv;

void execConfig(char* path) {
	execvp(path, myargv);
	fprintf(stderr, PROGRAMNAME ": the configured binary (%s)\n"
		"could not be executed: %s\n", path, strerror(errno));
	exit(1);
}

int main(int argc, char *argv[]) {
	char *progname;
	char *bsnam;

	progname = copyBuffer(argv[0]);
	bsnam = basename(progname);
	if (!strcmp(bsnam, PROGRAMNAME)) {
		/* pvswitch itself was called */
		int conferr;
		struct configrecord** cfg = readConfig(NULL);
		checkConfig(cfg);
		conferr = printErrors(cfg);
		if (! conferr) {
			fprintf(stderr, PROGRAMNAME ": config %s OK\n", CONFIGFILE);
		}
		freeConfig(cfg);
		copynotice();
	} else if (strendcmp("ver", bsnam)) {
		/* <whatever>ver has been called */
		struct configrecord** cfg;
		int conferr;
		bsnam[strlen(bsnam) - 3] = 0;
		cfg = readConfig(bsnam);
		if (cfg == NULL) {
			fprintf(stderr, "%sver: no " PROGRAMNAME " configuration for %s in " CONFIGFILE ".\n", bsnam, bsnam);
			exit(1);
		}
		checkConfig(cfg);
		conferr = printErrors(cfg);
		if (conferr) {
			fprintf(stderr, PROGRAMNAME ": config %s contains %d errors for %s\n", CONFIGFILE, conferr, bsnam);
			exit(1);
		} else {
			if (argc > 1) {
				/* argument driven */
				if (argc == 2) {
					configureVersion(cfg, argv[1]);
				} else {
					fprintf(stderr, "usage: %sver [version]\n", bsnam);
					fprintf(stderr, "   version - which version of %s should be used\n", bsnam);
					fprintf(stderr, "available %s versions:\n", bsnam);
					printConfig(cfg);
				}
			} else {	
				/* interactive selection */
				char buffer[TMP_BUFFER_SZ];
				char* ver;
				int rv;
	
				ver = buffer;
	
				fprintf(stderr, PROGRAMNAME ": choose a %s version:\n", bsnam);
				printConfig(cfg);
				rv = scanf("%s", buffer);	
				if (rv != 1) {
					/* on failure, do not change the configuration */
					ver = getEnvVersion(bsnam);
					fprintf(stderr, "%sver: invalid input\n", bsnam);
				}
				configureVersion(cfg, ver);
			}
		}
		freeConfig(cfg);	
	} else {
		/* the real execution must be done */
		char *ver;
		struct configrecord** cfg;
		struct configrecord* cfgrec;
		int conferr;

		myargv = argv;

		cfg    = readConfig(bsnam);
		conferr = checkConfig(cfg);
		
		if (conferr) {
			fprintf(stderr, PROGRAMNAME ": config %s contains %d errors for %s\n", CONFIGFILE, conferr, bsnam);
			exit(1);
		} else {
			ver = getEnvVersion(cfg[0]->part[PART_APROGRAM]);
	 		cfgrec = getConfig(cfg, ver);

			if (cfgrec == NULL) {
				fprintf(stderr, PROGRAMNAME ": no valid configuration for %s in %s\n", bsnam, CONFIGFILE);
				freeConfig(cfg);	
			} else {
				char* path = copyBuffer(cfgrec->part[PART_PATH]);
				freeConfig(cfg);	
				execConfig(path);
			}
		}
		free(progname);
		return(1);
	}
	return 0;
}
